#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 Module to provide class hierachy to simplify access to the BloxOne APIs

 Date Last Updated: 20230830

 Todo:

 Copyright (c) 2020 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
import bloxone
import logging
import json

# ** Global Vars **
__version__ = '0.9.2'
__author__ = 'Chris Marrison'
__email__ = 'chris@infoblox.com'
__doc__ = 'https://python-bloxone.readthedocs.io/en/latest/'
__license__ = 'BSD'


class b1platform(bloxone.b1oph):
    '''
    Class now reused for BloxOne Platform Methods, e.g. Audit Log
    Management of BloxOne On Prem Hosts is via the b1oph Class 
    b1oph class  is inherited here for compatibility
    '''

    def auditlog(self, **params):
        '''
        Get the audit log

        Parameters:
            **params (dict): Generic API parameters
        Returns:
            audit_log (list); list of dict
        '''
        # Local variables
        audit_log = []

        url = self.base_url + '/api/auditlog/' + self.api_version +'/logs'
        url = self._add_params(url, **params)

        logging.debug("URL: {}".format(url))

        # Call API
        response = self._apiget(url)

        if response.status_code in self.return_codes_ok:
            if 'results' in response.json().keys():
                audit_log = response.json()['results']
        else:
            audit_log = response.json()
        
        return audit_log


    def get_full_auditlog(self, **params):
        '''
        '''
        all_logs = []
        offset = 0
        limit = 1000

        audit_log = self.auditlog(_offset=str(offset), 
                                  _limit=str(limit), **params)
        while isinstance(audit_log, list) and len(audit_log):
            all_logs += audit_log
            offset += limit 
            audit_log = self.auditlog(_offset=str(offset), 
                                      _limit=str(limit), **params)

        return all_logs
    
    # Methods for undocumented API Calls - use at own risk

    def get_current_user(self, **params):
        '''
        Get Current User Data

        Parameters:
            **params (dict): Generic API parameters
        
        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.base_url + '/v2/current_user'
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiget(url)

        return response


    def get_current_user_accounts(self, **params):
        '''
        Get Current Users Accounts Data

        Parameters:
            **params (dict): Generic API parameters
        
        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.base_url + '/v2/current_user/accounts'
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiget(url)

        return response


    def get_current_tenant(self, **params):
        '''
        Get name of current tenant

        Parameters:
            **params (dict): Generic API parameters
        
        Returns:
            string containing name of tenant or '' on failure
        '''
        tenant_name = ''

        current_user = self.get_current_user()
        if current_user.status_code in self.return_codes_ok:
            user_id = current_user.json()['result']['account_id']
            user_accounts = self.get_current_user_accounts()
            for account in user_accounts.json()['results']:
                if user_id == account['id']:
                    tenant_name = account['name']
                    break
        else:
            logging.error('Failed to get account details API reponded with')
            logging.error(f'HTTP code: {current_user.status_code} ')
            logging.error(f'Response: {current_user.text}')
            tenant_name = ''
        
        return tenant_name

    
    def get_users(self, **params):
        '''
        Get User Data

        Parameters:
            **params (dict): Generic API parameters
        
        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.base_url + '/v2/users'
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiget(url)

        return response


    def audit_users(self, domains=[], ignore_service_accounts=True):
        '''
        Audit User Data for non compliant email domains

        Parameters:
            domain (list): List of valid email domains

        Returns:
            List of User Data (json)
        '''
        users = []
        user_data = []
        valid = False

        # Check for list of domains, otherwise use current users
        if len(domains) == 0:
            current_user = self.get_current_user().json()
            email = current_user['result']['email']
            domain = email.split('@')[1]
            domains.append(domain)
        logging.debug(f'Checking against domains: {domains}')

        # Get user data and check against valid domains
        response = self.get_users()
        user_data = response.json()['results']
        for user in user_data:
            valid = False
            if ignore_service_accounts and user['type'] == 'service':
                valid = True
            else:
                # Iterate through domains to check validity
                for domain in domains:
                    if domain in user['email']:
                        valid = True
                        break
            # Build list of exceptions
            if not valid:
                users.append(user)
        
        return users
    

    def get_user_id(self, email="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            email (str):    email address of user account
            include_path (bool): Include path to object id

        Returns:
            id (str):   object id or ""
        '''

        # Local Variables
        id = ""

        # Make API Call
        response = self.get_users()

        # Process response
        if response.status_code in self.return_codes_ok:
            users = response.json().get('results')
            if users:
                for user in users:
                    if email == user.get('email'):
                        id = user.get('id')
                        if not include_path:
                            id = id.rsplit('/',1)[1]
                        break 
                else:
                    logging.debug("User account {} not found."
                                  .format(email))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id


    def create_user(self, name='', 
                          email='', 
                          type='interactive',
                          authenticator='IDP',
                          groups=['user', 'act_admin'],
                          **params):
        '''
        Create User Account

        Parameters:
            **params (dict): Generic API parameters
        
        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.base_url + '/v2/users'
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        groups = self.get_group_ids(groups=groups)

        body = { 'name': name,
                 'email': email,
                 'type': type,
                 'authenticator': authenticator,
                 'groups': groups 
               }

        # Make API Call
        response = self._apipost(url, body=json.dumps(body))

        return response


    def delete_user(self, email='', **params):
        '''
        Create User Account

        Parameters:
            **params (dict): Generic API parameters
        
        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.base_url + '/v2/users'
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        user_id = self.get_id(url, key='email', 
                              value=email, 
                              include_path=False)
        
        logging.debug("ID: {}".format(user_id))

        if user_id:
            url = url + '/' + user_id
            # Make API Call
            response = self.delete(url)

        else:
            response = self._not_found_response

        return response


    def get_groups(self, **params):
        '''
        Get User Groups

        Parameters:
            **params (dict): Generic API parameters
        
        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.base_url + '/v2/groups'
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiget(url)

        return response


    def get_group_ids(self, groups=['user'], **params):
        '''
        Get User Group IDs

        Parameters:
            groups (list): List of group names
            **params (dict): Generic API parameters
        
        Returns:
            response object: Requests response object
        '''
        group_ids = []
        
        # Make API Call
        response = self.get_groups(_fields='name,id')
        if response.status_code in self.return_codes_ok:
            grps = response.json().get('results') 
            group_ids = [ grps[x].get('id') for x in range(len(grps)) 
                            if grps[x].get('name') in groups ]
            '''
            for grp in grps:
                if grp.get('name') in groups:
                    group_ids.append(grp.get('id'))
                    logging.debug(f'Found id for group {grp.get("name")}')
            '''
        else:
            logging.error('Failed to get User Groups API reponded with')
            logging.error(f'HTTP code: {response.status_code} ')
            logging.error(f'Response: {response.text}')
            group_ids = []

        return group_ids
    

    def global_search_version(self):
        '''
        Get the global search version information

        Returns:
            response object: Requests response object

        '''
        url = f'{self.base_url}/atlas-search-api/v1/version'
        return self.get(url)


    def global_search(self, query='', limit=None, filters=[], offset=0, **params):
        '''
        Perform a global search

        Parameters:
            query (str): Search term
            limit (int): Max number of records to match
            filters (list): list of filter definitions
            offset (int): record offset
            **params: Additional parameters
        
        Returns:
            response object: Requests response object

        '''
        body = {}
        url = f'{self.base_url}/atlas-search-api/v1/search'
        body['query'] = query
        body['filters'] = filters
        if limit:
            body['limit'] = limit
        if offset > 0:
            body['offset'] = offset
        if params:
            body.update(params)
        
        return self.post(url, body=json.dumps(body))


    def simple_global_search(self, query=''):
        '''
        Simple form of global_search() returning list of objects with ids

        Parameters:
            query (str): Search term
        
        Returns:
            result (list): List of dict of object ids and metadata
                           [ { '_id': id, 'metadata': dict of metadata } ]

        '''
        result = []
        response = self.global_search(query=query)
        if response.status_code in self.return_codes_ok:
            hits = response.json()['hits']['hits']
            if hits:
                for obj in hits:
                    result.append({ '_id': obj.get('_id'),
                                    'name': obj.get('_source').get('doc').get('name'),
                                    'tags': obj.get('_source').get('doc').get('tags'),
                                    'summary': obj.get('_source').get('doc').get('summary'),
                                    'metadata': obj.get('_source').get('metadata') })
            else:
                result = []
        else:
            logging.error('Global Search Failed')
            logging.error(f'HTTP code: {response.status_code} ')
            logging.error(f'Response: {response.text}')

        return result


    def get_join_token(self, name='', **params):
        '''
        Get Join Tokens

        Parameters:
            name (str): Optional name of JT

        Returns:
            response object: Requests response object
        '''
        url = f'{self.ztp_url}/jointoken'
        if name:
            url = self._add_params(url, _filter=f'name=="{name}"')
            url = self._add_params(url, first_param=False, **params)
        else:
            url = self._add_params(url, **params)

        return self._apiget(url)
    

    def get_join_token_id(self, name=''):
        '''
        Get the ID of a Join Token by name

        Parameters:
            name (str): Name of the JT
        
        Returns:
            id (str): id of the join token or ""
        '''
        response = self.get_join_token(name=name)
        if response.status_code in self.return_codes_ok:
            results = response.json().get('results')
            if results:
                id = results[0].get('id')
                id = id.rsplit('/',1)[1]
            else:
                id = ""
                logging.warning("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id


    def create_join_token(self, name='', desc='', tags={}):
        '''
        Create a new join token

        Parameters:
            name (str): Name of JT
            desc (str): Description
            tages (dict): Dictionary of tags to add

        Returns:
            response object: Requests response object
        '''
        body = {}
        url = f'{self.ztp_url}/jointoken'
        body = { 'name': name,
                 'description': desc }
        if tags:
            body['tags'] = tags

        logging.debug(f'Request body: {body}')

        return self._apipost(url, body=json.dumps(body))


    def revoke_join_token(self, name=''):
        '''
        Revoke a join token by name

        Parameters:
            name (str): Name of JT
        
        Returns:
            response object: Requests response object
        '''
        id = self.get_join_token_id(name=name)
        if id:
            url = f'{self.ztp_url}/jointoken/{id}'
            response = self._apidelete(url)
        else:
            response = self._not_found_response()

        return response


    def revoke_join_tokens(self, id_list=[]):
        '''
        Revoke a set of join tokens

        Parameters:
            id_list (list): list of JT ids
        
        Returns:
            response object: Requests response object
        '''
        body = {}

        if isinstance(id_list, list):
            url = f'{self.ztp_url}/jointokens'
            body = { "ids": id_list }
            response = self._apidelete(url, body=json.dumps(body))
        else:
            logging.warning('id_list was not a list')
            response = self._not_found_response()

        return response


#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 BloxOne DDI API Wrapper Class and Helpers

 Author: Chris Marrison

 Date Last Updated: 20230413

 Todo:

 Copyright (c) 2020 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
__version__ = '0.2.6'
__author__ = 'Chris Marrison/Krishna Vasudevan'
__author_email__ = 'chris@infoblox.com'

import bloxone
import logging
import json
from requests.models import Response

# ** Global Vars **

class b1tdc(bloxone.b1):
    '''
    BloxOne ThreatDefence Cloud API Wrapper
    '''

    # Generic Methods
    def get(self, objpath, id="", action="", **params):
        '''
        Generic get object wrapper for Threat Defense Cloud

        Parameters:
            objpath (str):  Swagger object path
            action (str):   Optional object action
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tdc_url + objpath
        if id:
            url = self._use_obj_id(url, id=id)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response


    def post(self, objpath, body=""):
        '''
        Generic create object wrapper for Threat Defense Cloud

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tdc_url + objpath

        # Make API Call
        response = self._apipost(url, body)

        return response


    def create(self, objpath, body=""):
        '''
        Generic create object wrapper for Threat Defense Cloud

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tdc_url + objpath

        # Make API Call
        response = self._apipost(url, body)

        return response


    def delete(self, objpath, id="", body=""):
        '''
        Generic delete object wrapper for Threat Defense Cloud

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Object id to delete
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tdc_url + objpath
        if id:
            url = self._use_obj_id(url, id=id)

        # Make API Call
        response = self._apidelete(url, body)

        return response

       
    def put(self, objpath, id="", body=""):
        '''
        Generic put object wrapper for Threat Defense Cloud

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Object id to update
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tdc_url + objpath
        url = self._use_obj_id(url, id=id)

        # Make API Call
        response = self._apiput(url, body)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for Threat Defense Cloud

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tdc_url + objpath
        url = self._use_obj_id(url, id=id)

        # Make API Call
        response = self._apipatch(url, body)

        return response

 
   # *** Helper Methods ***

    def get_id(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match

        Returns:
            id (str):   object id or ""
        '''

        # Local Variables
        id=""

        # Build url
        url = self.tdc_url + objpath
        url = url + '?_fields=' + key + ',id'

        # Make API Call
        response = self._apiget(url)

        # Process response
        if response.status_code in self.return_codes_ok:
            objs = json.loads(response.text)
            # Look for results
            if "results" in objs.keys():
                for obj in objs['results']:
                    if obj[key] == value:
                        id = obj['id']
                        if not include_path and "/" in str(id):
                            id = id.rsplit('/',1)[1]
                if not id:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))
        logging.debug("id: {}".format(id)) 

        return id


    def get_object_by_key(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match

        Returns:
            id (str):   object id or ""
        '''
        return self.get(objpath, id=self.get_id(objpath, key=key, value=value))


    def get_custom_lists(self, **params):
        '''
        Get custom lists

        Parameters:
            Additional parameters per API documentation
             
        Returns:
            response object: Requests response object
        '''
        url = self.tdc_url + '/named_lists'
        url = self._add_params(url, **params)

        # Make API Call
        response = self._apiget(url)

        return response


    def get_custom_list(self, name='', **params):
        '''
        Get the named custom list

        Parameters:
            name (str): Name of custom list
        
        Returns:
            response object: Requests response object
        '''
        id = self.get_id('/named_lists', key='name', value=name)
        if id:
            logging.debug(f'Custom list: {name} with id {id} found.')
            response = self.get('/named_lists', id=id, **params)
        else:
            logging.debug(f'Custom list: {name} not found.')
            response = self._not_found_response('named_list')

        return response


    def create_custom_list(self, name='', 
                                 confidence='HIGH', 
                                 items=[], 
                                 items_described=[]):
        '''
        Create deny custom lists

        Parameters:
            name (str): Name of custom list
            confidence (str): Confidence level
            items (list): List of indicators
            items_described (list): List of {"description": "", "item": ""}
        
        Note: 
            if items and items_described are both included items_described 
            will take precidence.

        Returns:
            response object: Requests response object
        '''
        if items:
            body = { "name": name,
                     "type": "custom_list",
                     "confidence_level": confidence,
                     "items": items }
        elif items_described:
            body = { "name": name,
                     "type": "custom_list",
                     "confidence_level": confidence,
                     "items_described": items_described }
        else:
            body = { "name": name,
                     "type": "custom_list",
                     "confidence_level": confidence }

        logging.debug("Body:{}".format(body))
        response = self.create('/named_lists', body=json.dumps(body))

        return response


    def add_items_to_custom_list(self, name='', items=[], items_described=[]):
        '''
        Add items to an existing custom list

        Parameters:
            name (str): Name of custom list
            items (list): List of indicators
            items_described (list): List of {"description": "", "item": ""}
        
        Note: 
            For updates both items and items_described can be included.

        Returns:
            response object: Requests response object
        '''
        body = {}
        id = self.get_id('/named_lists', key='name', value=name)
        if id:
            logging.debug(f'Custom list: {name} with id {id} found.')
            request = f'/named_lists/{id}/items'
            if items:
                body = { "items": items }
            if items_described:
                body.update( { "items_described": items_described } )

            logging.debug("Body:{}".format(body))
            response = self.post(request, body=json.dumps(body))
        else:
            logging.debug(f'Custom list: {name} not found.')
            response = self._not_found_response('named_list')

        return response


    def delete_items_from_custom_list(self, name='', 
                                            items=[], 
                                            items_described=[]):
        '''
        Delete items to an existing custom list

        Parameters:
            name (str): Name of custom list
            items (list): List of indicators
            items_described (list): List of {"description": "", "item": ""}
        
        Note: 
            For updates both items and items_described can be included.

        Returns:
            response object: Requests response object
        '''
        body = {}
        id = self.get_id('/named_lists', key='name', value=name)
        if id:
            logging.debug(f'Custom list: {name} with id {id} found.')
            request = f'/named_lists/{id}/items'
            if items:
                body = { "items": items }
            if items_described:
                body.update( { "items_described": items_described } )

            logging.debug("Body:{}".format(body))
            response = self.delete(request, body=json.dumps(body))
        else:
            logging.debug(f'Custom list: {name} not found.')
            response = self._not_found_response('named_list')

        return response

 
    def delete_custom_lists(self, names=[]):
        '''
        Delete custom list
            
        Parameters:
            name (list): List of names(str) to delete
        
        Returns:
            response object: Requests response object or None
        '''
        ids = []
        for name in names:
            id = self.get_id('/named_lists', key="name", value=name)
            if id:
                logging.debug(f'Custom list: {name} with id {id} found.')
                ids.append(str(id))
            else:
                logging.debug(f'Custom list: {name} not found.')
        if ids:
            body = { 'ids': ids }
            logging.debug("Body:{}".format(body))
            response = self.delete('/named_lists', body=json.dumps(body))
        else:
            logging.warning('No custom lists found')
            response = self._not_found_response('named_list')
        
        return response


#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 BloxOne DDI API Wrapper Class and Helpers

 Author: Chris Marrison

 Date Last Updated: 20240701

 Todo:

 Copyright (c) 2020 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
__version__ = '0.4.8'
__author__ = 'Chris Marrison'
__author_email__ = 'chris@infoblox.com'

import bloxone
import logging
import json

# ** Global Vars **

class b1ddi(bloxone.b1):
    '''
    BloxOne DDI API Wrapper Class
    '''

    # Generic Methods
    def get(self, objpath, id="", action="", **params):
        '''
        Generic get object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.ddi_url + objpath
        url = self._use_obj_id(url, id=id, action=action)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response


    def post(self, objpath, id="", action="", body="", **params):
        '''
        Generic POST object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"
            body (str):     JSON formatted data payload
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.ddi_url + objpath
        url = self._use_obj_id(url, id=id, action=action)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apipost(url, body)

        return response


    def create(self, objpath, body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.ddi_url + objpath
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipost(url, body)

        return response


    def delete(self, objpath, id="", recursive_delete=False, **params):
        '''
        Generic delete object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Object id to delete

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.ddi_url + objpath
        url = self._use_obj_id(url,id=id)
        if recursive_delete:
            options = { '_options': 'recurse=true' }
            url = self._add_params(url, **options)
            url = self._add_params(url, first_param=False, **params)
        else:
            url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apidelete(url)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.ddi_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response


    def replace(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.ddi_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipatch(url, body)

        return response


   # *** Helper Methods ***

    def get_id(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match
            include_path (bool): Include path to object id

        Returns:
            id (str):   object id or ""
        '''

        # Local Variables
        id = ""
        filter = key+'=="'+value+'"'
        fields = key + ',id'

        # Make API Call
        response = self.get(objpath, _filter=filter, _fields=fields)

        # Process response
        if response.status_code in self.return_codes_ok:
            obj = response.json()
            # Look for results
            if "results" in obj.keys():
                obj = obj['results']
                if obj:
                    id = obj[0]['id']
                    if not include_path:
                        id = id.rsplit('/',1)[1]
                else:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id


    def search_response(self, response, key="", value="", include_path=False):
        '''
        Get object id using key/value pair by searching a 
        Request response object.

        Parameters:
            response object:     Request response obj
            key (str):          name of key to match
            value (str):        value to match
            include_path (bool): Include path to object id

        Returns:
            id (str):   object id or ""
        '''

        # Process response
        id = ""
        if response.status_code in self.return_codes_ok:
            objs = json.loads(response.text)
            # Look for results
            if "results" in objs.keys():
                for obj in objs['results']:
                    if obj[key] == value:
                        id = obj['id']
                        if not include_path:
                            id = id.rsplit('/',1)[1]
                if not id:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))
        logging.debug("id: {}".format(id)) 

        return id


    def get_object_by_key(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match

        Returns:
            id (str):   object id or ""
        '''
        response = None
        id=self.get_id(objpath, key=key, value=value)
        if id:
           response = self.get(objpath, id=id) 
        else:
            response = self._not_found_response

        return response


    def get_option_ids(self, option_space=""):
        '''
        Return a dictionary of DHCP Option IDs
        Based on idea/code from John Neerdael

        Parameters:
            option_space (str): Optional Option Space ID
        
        Returns:
            option_ids (dict): Dictionary keyed on option number of ids
        '''
        option_ids = {}

        response = self.get('/dhcp/option_code', _fields='code,id')
        if response.status_code in self.return_codes_ok:
            if 'results' in response.json().keys():
                optioncodes = response.json()['results']
                for item in optioncodes:
                    code = item['code']
                    id = item['id']
                    option_ids.update({ code: id })
        
        return option_ids


    def get_tags(self, objpath, id=""):
        '''
        Get tags for an object id

        Parameters:
            objpath (str):  Swagger object path

            id (str): id of object

        Returns:
            tags (dict): Dictionary of current tags
                         or empty dict if none
        
        .. todo::
            * make generic, however, this requires the below
            * Lookup dictionary of 'required fields' per object type
        '''
        tags = {}
        response = self.get(objpath, id=id, _fields="tags")
        if response.status_code in self.return_codes_ok:
            tags = json.loads(response.text)
            tags = tags['result']
        else:
            tags = {}
        
        return tags


    def add_tag(self, objpath, id, tagname="", tagvalue=""):
        '''
        Method to add a tag to an existing object
        Note: PUT/update Not Implemented in API as yet

        Parameters:
            objpath (str):  Swagger object path
            id (str): Object ID
            tagname (str): Name of tag to add
            tagvalue (str): Value to associate with tag

        Returns:
            response object: Requests response object
        '''
        response = self.get(objpath, id=id, _fields='tags')
        if response.status_code in self.return_codes_ok:
            if 'result' in response.json().keys():
                data = response.json()['result']
            elif 'results' in response.json().keys():
                data = response.json()['results']
            logging.debug("Existing tags: {}".format(data.get('tags')))
        else:
            data = {}

        if data:
            # Add new tag to data
            if tagname:
                data['tags'].update({tagname: tagvalue})
                logging.debug("New tags: {}".format(data.get('tags')))
            # Update object
            response = self.replace(objpath, id=id, body=json.dumps(data))
        else:
            logging.debug("No object found")

        return response


    def delete_tag(self, objpath, id="", tagname=""):
        '''
        Method to delete a tag from an existing On Prem Host

        Parameters:
            objpath (str):  Swagger object path
            tagname (str): Name of tag to add

        Returns:
            response object: Requests response object
        '''
        # tags = self.get_tags('/on_prem_hosts', id=id)
        response = self.get(objpath, id=id, _fields='tags')
        if response.status_code in self.return_codes_ok:
            data = response.json()['result']
            logging.debug("Existing tags: {}".format(data['tags']))
            # Delete tag from data
            if tagname in data['tags'].keys():
                data['tags'].pop(tagname, True)
                logging.debug("New tags: {}".format(data['tags']))
                # Update object
                response = self.replace(objpath, id=id, body=json.dumps(data))

        return response


    # *** Undocumented DNS Calls ***

    def get_zone_child(self, parent="zone", name="", id="", fields=""):
        '''
        Method to retrieve Zones in specified view

        Parameters:
            name (str): BloxOne object id
            id (str):   BloxOne object id
            **params (dict): Generic API parameters

        Returns:
           response object: Requests response object
        '''
        if name and not id:
            if parent == "view":
                id = self.get_id('/dns/view', key="name", value=name, include_path=True)
                # result = self.get('/dns/view', _fields="name,id")
                # logging.debug(result.text)
            elif parent == "zone":
                id = self.get_id('/dns/auth_zone', key="fqdn", value=name, include_path=True)
                # result = self.get('/dns/auth_zone', _fields="fqdn,id")
                # logging.debug(result.text)
            logging.debug("id: {}".format(id)) 
        # Build URL
        objpath = '/dns/zone_child' + '?_filter=parent=="' + id + '"'
        if fields:
            objpath = objpath + '&_fields=' + fields
        logging.debug("Objectpath: {}".format(objpath))
        
        # Call BloxOne API GET Method
        return self.get(objpath)
        



#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:
    Simple utility functions for data type validation, domain handling,
    and data normalisationa specifically with the aim of supporting
    queries to TIDE and Dossier.

 Requirements:
   Python3 with re, ipaddress, requests 

 Author: Chris Marrison

 Date Last Updated: 20210714

 Todo:

 Copyright (c) 2018 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
__version__ = '0.0.3'
__author__ = 'Chris Marrison'
__author_email__ = 'chris@infoblox.com'

import logging
import os
import re
import ipaddress
import urllib.parse
import sqlite3

# ** Global Vars **


# ** Data Validation functions ** #

def data_type(qdata, host_regex, url_regex):
    '''
    Validate and determine data type (host, ip or url)

    Parameters:
        qdata (str): data to determine type/validity
        host_regex/url_regex (re): pre-compiled regexes

    Returns:
        dtype (str): data type of qdata ("ip", "host", or "url")
    '''
    if validate_ip(qdata):
        dtype = "ip"
    elif validate_url(qdata, url_regex):
        dtype = "url"
    elif validate_fqdn(qdata, host_regex):
        dtype = "host"
    else:
        dtype = "invalid"
    return dtype


def buildregex():
    '''
    Pre-compile 'standard' regexes as used by data_type and
    validate_XXX functions

    Parameters:
        none

    Returns:
        host_regex (re): Compiled regex for hostnames
        url_regex (re): Compiled regex for URLs
    '''
    # Added _ for support of Microsoft domains
    host_regex = re.compile(r'(?!-)[A-Z\d\-\_]{1,63}(?<!-)$', re.IGNORECASE)
    url_regex = re.compile(
        # r'^(?:http|ftp)s?://' # http:// or https://
        # http:// or https://
        r'^(?:http)s?://'
        # domain...
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'
        # localhost...
        r'localhost|'
        # ...or ipv4
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'
        # ...or ipv6
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'
        # optional port
        r'(?::\d+)?'
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    return host_regex, url_regex


def validate_fqdn(hostname, regex):
    '''
    Validate input data is a legitmate fqdn

    Parameters:
        hostname (str): fqdn as a string
        regex (obj): Compiled regex for hostnames

    Returns:
        bool: Return True for valid and False otherwise

    '''
    if len(hostname) > 255 or len(hostname) < 1:
        result = False
    if hostname.endswith("."):
        # strip exactly one dot from the right, if present
        hostname = hostname[:-1]
    result = all(regex.match(x) for x in hostname.split("."))

    return result


def validate_ip(ip):
    '''
    Validate input data is a valid IP address

    Parameters:
        ip (str): ip address as a string

    Returns:
        bool: Return True for valid and False otherwise

    '''
    try:
        ipaddress.ip_address(ip)
        result = True
    except ValueError:
        result = False
    return result


def validate_url(url, regex):
    '''
    Validate input data is a valid URL

    Parameters:
        url (str): string to verify as URL
        regex (re): pre-compiled regex obj

    Returns:
        bool: Return True for valid and False otherwise

    '''
    result = regex.match(url)
    return result


# ** Misc Functions **

def reverse_labels(domain):
    '''
    Reserve order of domain labels (or any dot separated data, e.g. IP)

    Parameters:
        domain (str): domain.labels
    Returns:
        rdomain (str): labels.domain
    '''
    rdomain = ""
    labels = domain.split(".")
    for l in reversed(labels):
        if rdomain:
            rdomain = rdomain+"."+l
        else:
            rdomain = l
    return rdomain


# ** Data Conversion functions ** #

def convert_url_to_host(url):
    '''
    Break down URL and return host element

    Parameters:
        url (str): Validated URL

    Returns:
        host (str): hostname or ip

    '''
    # Break down URL and return host element
    componants = url.split("/")
    host = componants[2]

    # Remove port if present
    if ":" in host:
        componants = host.split(":")
        host = componants[0]

    return host


def normalise(item, itype=None, port=False, www=False):
    '''
    Take ip, host or url item process and
    return normalise data.

    Parameters:
        item (str): item to normalise
        itype (str): One of ["host", "url", "ip"]
        port (bool): stip port number e.g. :8080
        www (bool): strip www. from hostname

    Returns:
        normalised (str): Normalised item or "invalid"

    '''
    # Local varlables

    # Check for itype being set
    if not itype:
        # Build regexes for data_type checking
        host_regex, url_regex = buildregex()
        itype = data_type(item, host_regex, url_regex)

    if itype != "invalid":
        # Normalise item
        if itype == "url":
            # Parse URL
            parsed_url = urllib.parse.urlparse(item)
            # Strip unneeded elements (protocol, params, query, *port, *www)
            if port:
                item = parsed_url.hostname + parsed_url.path
            else:
                item = parsed_url.netloc + parsed_url.path

            # Remove www if required
            if www and item.startswith('www.'):
                item = item[4:]

        elif itype == "host":
            # Remove trailing dot if present
            if item.endswith("."):
                item = item[:-1]

            # Remove port if required
            if port and ":" in item:
                componants = item.split(":")
                item  = componants[0]

            # Remove www if required
            if www and item.startswith('www.'):
                item = item[4:]

            # Ensure lower case
            item = item.lower()

        elif itype == "ip":
            # Remove port if required
            if port and ":" in item:
                componants = item.split(":")
                item = componants[0]

        normalised = item

    else:
        normalised = itype

    return normalised


def count_labels(fqdn):
    '''
    Count number of labels in an FQDN

    Parameters:
        fqdn (str): Hostname as fqdn

    Returns:
        count (int): number of labels
    '''

    labels = []

    labels = fqdn.split(".")
    count = len(labels)

    return count


def strip_host(fqdn):
    '''
    Take FQDN and strip first label
    or fqdn if no. of labels is 2 or less

    Parameters:
        fqdn (str): Hostname as fqdn

    Returns:
        domain (str): stripped domain down to two labels
    '''
    labels = []
    domain = ""

    labels = fqdn.split(".")
    if len(labels) <= 2:
        domain = fqdn
    else:
        domain = get_domain(fqdn, no_of_labels=(len(labels) - 1))
        '''
        Alternate code with get_domain

        host = labels.pop(0)
        for label in labels:
            domain += label
            domain += "."
        domain = domain.rstrip(".")
        '''

    return domain


def get_domain(fqdn, no_of_labels=2):
    '''
    Take FQDN and return n label domain
    or fqdn if no. of labels is 2 or less

    Parameters:
        fqdn (str): Hostname as fqdn
        no_of_labels (int): Number of labels to return
            default = 2

    Returns:
        domain (str): N label domain name or fqdn

    '''

    # Local variables
    labels = []
    domain = ""
    count = 0

    labels = fqdn.split(".")
    count = len(labels)

    if count <= no_of_labels:
        domain = fqdn
    else:
        for label in range((count - no_of_labels), (count)):
            domain += labels[label]
            domain += "." 
        # Strip last "."
        domain = domain.rstrip(".")

    return domain

# ** Local db Functions **

def opendb(dbfile):
    '''
    Open sqlite db and return cursor()

    Parameters:
        dbfile (str): path to file

    Returns:
        cursor (obj): Returns a db.cursor()
    '''
    if os.path.isfile(dbfile):
        db = sqlite3.connect(dbfile)
        db.row_factory = sqlite3.Row
        cursor = db.cursor()
    else:
        logging.error("Database "+dbfile+" not found.")
        cursor = None

    return cursor


def get_table(cursor):
    '''
    Determine db table and return

    Parameters:
        cursor (obj): db.cursor object

    Returns:
        table (str): Table name as string
    '''
    # ** Determine table name ** #
    select = 'SELECT name FROM sqlite_master WHERE type="table"'
    cursor.execute(select)
    tables = cursor.fetchall()
    # ** Expecting single table ** #
    if len(tables) == 1:
        # ** Assume table is correct (even if it isn't) ** #
        table = tables[0][0]
    else:
        # ** Print error and exit ###
        logging.error("DB not of correct format - too many tables")
        table = None

    return table


def db_query(db_cursor, table, query_type, query_data, *flags):
    '''
    Perform db query and return appropriate rows

    Parameters:
        db_cursor (obj): db.cursor object
        table (str): database table name
        query_type (str): data type for query
        query_data (str): search string

    Returns:
        rows (list): (All matching db rows
    '''
    if query_type == "host":
        # ** Form DB Query **
        select = ('SELECT * FROM '
                  + table + ' WHERE host="' + query_data
                  + '" OR domain="' + query_data+'"')
        # Ignore class = "Policy"
        if flags:
            select = select+' AND property!="Policy_UnsolictedBulkEmail"'
    elif query_type == "ip":
        # ** Form DB Query **
        select = 'SELECT * FROM '+table+' WHERE ip="'+query_data+'"'
        # Ignore class = "Policy"
        if flags:
            select = select+' AND property!="Policy_UnsolictedBulkEmail"'
    elif query_type == "url":
        # ** Form DB Query **
        select = 'SELECT * FROM '+table+' WHERE url="'+query_data+'"'
        # Ignore class = "Policy"
        if flags:
            select = select+' AND property!="Policy_UnsolictedBulkEmail"'
    else:
        logging.error("Invalid Type for {} data type for {}"
                      " - should not be here".format(query_type, query_data))
        return None

    db_cursor.execute(select)
    rows = db_cursor.fetchall()

    return rows



#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 Module to provide class hierachy to simplify access to the BloxOne APIs

 Date Last Updated: 20240405

 Todo:

    api_key format verification upon inifile read.

 Copyright (c) 2020-2023 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
import logging
import configparser
import requests
import os
import re

# ** Global Vars **
__version__ = '0.9.1'
__author__ = 'Chris Marrison'
__email__ = 'chris@infoblox.com'
__doc__ = 'https://python-bloxone.readthedocs.io/en/latest/'
__license__ = 'BSD'


# Custom Exceptions
class IniFileSectionError(Exception):
    '''
    Exception for missing section in ini file
    '''
    pass

class IniFileKeyError(Exception):
    '''
    Exception for missing key in ini file
    '''
    pass

class APIKeyFormatError(Exception):
    '''
    Exception for API key format mismatch
    '''
    pass

# ** Facilitate ini file for basic configuration including API Key

def read_b1_ini(ini_filename):
    '''
    Open and parse ini file

    Parameters:
        ini_filename (str): name of inifile

    Returns:
        config (dict): Dictionary of BloxOne configuration elements

    Raises:
        IniFileSectionError
        IniFileKeyError
        APIKeyFormatError
        FileNotFoundError

    '''
    # Local Variables
    cfg = configparser.ConfigParser()
    config = {}
    ini_keys = ['url', 'api_version', 'api_key']
 
    # Check for inifile and raise exception if not found
    if os.path.isfile(ini_filename):
        # Attempt to read api_key from ini file
        try:
            cfg.read(ini_filename)
        except configparser.Error as err:
            logging.error(err)

        # Look for BloxOne section
        if 'BloxOne' in cfg:
            for key in ini_keys:
                # Check for key in BloxOne section
                if key in cfg['BloxOne']:
                    config[key] = cfg['BloxOne'][key].strip("'\"")
                    logging.debug('Key {} found in {}: {}'
                                 .format(key, ini_filename, config[key]))
                else:
                    logging.error('Key {} not found in BloxOne section.'
                                 .format(key))
                    raise IniFileKeyError('Key "' + key + '" not found within' 
                        '[BloxOne] section of ini file {}'.format(ini_filename))
                    
        else:
            logging.error('No BloxOne Section in config file: {}'
                         .format(ini_filename))
            raise IniFileSectionError('No [BloxOne] section found in ini file {}'
                                     .format(ini_filename))
        
    else:
        raise FileNotFoundError('ini file "{}" not found.'.format(ini_filename))

    return config


def verify_api_key(apikey):
    '''
    Verify format of API Key
    
    Parameters:
       apikey (str): api key

    Returns:
        bool: True is apikey passes format validation
    '''
    if re.fullmatch('[a-z0-9]{32}|[a-z0-9]{64}', apikey, re.IGNORECASE):
        status = True
    else:
        status = False

    return status


class b1:
    '''
    Parent Class to simplify access to the BloxOne APIs for subclasses
    Can also be used to genericallly access the API

    Raises:
        IniFileSectionError
        IniFileKeyError
        APIKeyFormatError
        FileNotFoundError
    '''

    def __init__(self, cfg_file='config.ini', 
                       api_key='', 
                       url='https://csp.infoblox.com', 
                       api_version='v1'):
        '''
        Read ini file and set attributes

        Parametrers:
            cfg_file (str): Override default ini filename
            api_key (str): Use API Key instead of ini
            url (str): Override URL, applies only if api_key specified
            api_version (str): API version, applies only if api_key specified

        '''

        self.cfg = {}

        if api_key:
            self.api_key = api_key
            # Create base URLs
            self.base_url = url
            self.api_version = api_version

        else:
            # Read ini file
            self.cfg = read_b1_ini(cfg_file)
            self.api_key = self.cfg['api_key']
        
            # Create base URLs
            self.base_url = self.cfg['url']
            self.api_version = self.cfg['api_version']

        # Verify format of API Key
        if verify_api_key(self.api_key):
            logging.debug('API Key passed format verification')
        else:
            logging.debug('API Key {} failed format verification'
                          .format(self.api_key))
            raise APIKeyFormatError('API Key {} failed format verification'
                                    .format(self.api_key))

        # Define generic header
        self.headers = ( {'content-type': 'application/json',
                        'Authorization': 'Token ' + self.api_key} )
        

        # B1 & B1DDI URLs
        self.anycast_url = self.base_url + '/api/anycast/' + self.api_version
        self.authn_url = self.base_url + '/api/authn/' + self.api_version
        self.bootstrap_url = self.base_url + '/bootstrap-app/' + self.api_version
        self.cdc_url = self.base_url + '/api/cdc-flow/' + self.api_version
        self.diagnostics_url = self.base_url + '/diagnostic-service/' + self.api_version
        self.ddi_url = self.base_url + '/api/ddi/' + self.api_version
        self.host_url = self.base_url + '/api/host_app/' + self.api_version
        self.infra_url = self.base_url + '/api/infra/' + self.api_version
        self.monitoring_url = self.base_url + '/api/monitoring/' + self.api_version
        self.notifications_url = self.base_url + '/atlas-notifications-config/'+ self.api_version
        self.ntp_url = self.base_url + '/api/ntp/' + self.api_version
        self.sw_url = self.base_url + '/api/upgrade_policy/' + self.api_version
        self.ztp_url = self.base_url + '/atlas-host-activation/' + self.api_version
        
        # B1TD URLs
        self.tdc_url = self.base_url + '/api/atcfw/' + self.api_version
        self.tdep_url = self.base_url + '/api/atcep/' + self.api_version
        self.tddfp_url = self.base_url + '/api/atcdfp/' + self.api_version
        self.tdlad_url = self.base_url + '/api/tdlad/' + self.api_version
        self.tide_url = self.base_url + '/tide/api' 
        self.dossier_url = self.base_url + '/tide/api/services/intel/lookup'
        self.threat_enrichment_url = self.base_url + '/tide/threat-enrichment'
        self.insights_url = self.base_url + '/api/' + self.api_version

        # Reporting URLs (undocumented - subject to change)
        self.ti_reports_url = self.base_url + '/api/reports/' + self.api_version
        self.aggr_reports_url  = self.ti_reports_url + '/activity/aggregations'
        self.sec_act_url = self.base_url + '/api/ti-reports/v1/activity/hits'

        # List of successful return codes
        self.return_codes_ok = [200, 201, 204]

        return


    def _add_params(self, url, first_param=True, **params):
        # Add params to API call URL
        if len(params):
            for param in params.keys():
               if first_param:
                   url = url + '?'
                   first_param = False
               else:
                   url = url + '&'
               url = url + param + '=' + str(params[param])
        
        return url


    def _apiget(self, url):    
     # Call BloxOne API
        try:
            response = requests.request("GET",
                                        url,
                                        headers=self.headers)
        # Catch exceptions
        except requests.exceptions.RequestException as e:
            logging.error(e)
            logging.debug("url: {}".format(url))
            raise

        # Return response code and body text
        # return response.status_code, response.text
        return response


    def _apipost(self, url, body, headers=""):    
        # Set headers
        if not headers:
            headers = self.headers
     
        # Call BloxOne API
        try:
            response = requests.request("POST",
                                        url,
                                        headers=headers,
                                        data=body)
        # Catch exceptions
        except requests.exceptions.RequestException as e:
            logging.error(e)
            logging.debug("url: {}".format(url))
            logging.debug("body: {}".format(body))
            raise

        # Return response code and body text
        return response

 
    def _apidelete(self, url, body=""):    
     # Call BloxOne API
        try:
            response = requests.request("DELETE",
                                        url,
                                        headers=self.headers,
                                        data=body)
        # Catch exceptions
        except requests.exceptions.RequestException as e:
            logging.error(e)
            logging.debug("URL: {}".format(url))
            logging.debug("BODY: {}".format(body))
            raise

        # Return response code and body text
        return response


    def _apiput(self, url, body):    
     # Call BloxOne API
        try:
            response = requests.request("PUT",
                                        url,
                                        headers=self.headers,
                                        data=body)
        # Catch exceptions
        except requests.exceptions.RequestException as e:
            logging.error(e)
            logging.debug("url: {}".format(url))
            logging.debug("body: {}".format(body))
            raise

        # Return response code and body text
        return response

 
    def _apipatch(self, url, body):    
        # Call BloxOne API
        try:
            response = requests.request("PATCH",
                                        url,
                                        headers=self.headers,
                                        data=body)
        # Catch exceptions
        except requests.exceptions.RequestException as e:
            logging.error(e)
            logging.debug("url: {}".format(url))
            logging.debug("body: {}".format(body))
            raise

        # Return response code and body text
        return response


    def _use_obj_id(self, url, id="", action=""):
        '''
        Update URL for use with object id
        
        Parameters:
            id (str): Bloxone Object id
            action (str): e.g. nextavailableip

        Returns:
            string : Updated url
        '''
        # Check for id and next available IP
        if id:
            url = url + '/' + str(id)
            if action:
                url = url + '/' + action
        else:
            if action:
                logging.debug("Action {} not supported without " 
                              "a specified object id.")
        
        return url
    
    
    def _not_found_response(self, b1object='object'):
        '''
        Generate a response object without an API call

		Parameters:
            b1object (str): Name of object to use in error

        Returns:
            requests response object
        
        '''
        err_msg = f'{{"error":[{{"message":"{b1object} not found"}}]}}'
        response = requests.Response()
        response.status_code = 400
        response._content = str.encode(err_msg)

        return response


    # Public Generic Methods
    def get(self, url, id="", action="", **params):
        '''
        Generic get object wrapper 

        Parameters:
            url (str):      Full URL
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self._use_obj_id(url, id=id, action=action)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response


    def post(self, url, id="", action="", body="", **params):
        '''
        Generic Post object wrapper 

        Parameters:
            url (str):      Full URL
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self._use_obj_id(url, id=id, action=action)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apipost(url, body)

        return response


    def create(self, url, body=""):
        '''
        Generic create object wrapper 

        Parameters:
            url (str):  Full URL
            body (str): JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipost(url, body)

        return response


    def delete(self, url, id="", body=""):
        '''
        Generic delete object wrapper

        Parameters:
            url (str):  Full URL
            id (str):   Object id to delete
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        if id:
            url = self._use_obj_id(url,id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apidelete(url, body=body)

        return response


    def update(self, url, id="", body=""):
        '''
        Generic create object wrapper 

        Parameters:
            url (str):  Full URL
            body (str): JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url if needed
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response


    def replace(self, url, id="", body=""):
        '''
        Generic create object wrapper 

        Parameters:
            url (str):  Full URL
            body (str): JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipatch(url, body)

        return response